/*
 * Copyright (C) 2003 ichitokumei.
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <qfile.h>
#include "inkwpdata.h"

#define SPACE_WIDTH 0.5
#define ENTER_WIDTH 0.5
#define END_WIDTH 0.5

#define MAGIC ("INKWP")
#define VER (0)

/*!
  \class InkWpChar inkwpdata.h

  A class for characters in InkWp.

  There are two types of charcters: 
    Handwriting:
      a hand-written character, represented by 
	  (1) a QPointArray \c points(),
	  (2) a QBitArray \c connected(), 
	  which tells if the point is connected with the last point. 
	  (3) an integer \c standardHeight, 
	  which tells the standard height when it was input. 
	  
	  When drawn, 
	  (a) if the height of the handwriting >= standardHeight, 
	  the character is scaled to the given height. 
	  (b) if not, it is scaled relative to the standardHeight.
    String:
	  A QString. Currently, the following are used: 
	  "\n": enter
	  " " : space
	  "" : end of document(for later extention)
 */

/*!
  Default constructor.
 */
InkWpChar::InkWpChar() {
}
/*!
  Constructs an InkWpChar with type t.
 */
InkWpChar::InkWpChar(Type t) : type_(t) {
}
/*!
  Constructs an InkWpChar with type Handwriting, standardHeight = i.
 */
InkWpChar::InkWpChar(int i) : type_(Handwriting), standardHeight_(i) {
}
/*!
  Constructs an InkWpChar with type String, test = s.
 */
InkWpChar::InkWpChar(QString s) : type_(String), text_(s) {
}

void InkWpChar::setType(Type t) {
  type_ = t;
}
InkWpChar::Type InkWpChar::type() const {
  return type_;
}
void InkWpChar::setStandardHeight(int i) {
  standardHeight_ = i;
}
int InkWpChar::standardHeight() const {
  return standardHeight_;
}

int InkWpChar::width(int h) {
  int w = 0;
  switch(type()) {
  case String: 
	if (text() == " ") 
	  w = h * SPACE_WIDTH;
	else if (text() == "\n") 
	  w = h * ENTER_WIDTH;
	else if (text() == "")
	  w = h * END_WIDTH;
	break;
  case Handwriting: 
	{
	  QRect r = points().boundingRect();
	  int hh = QMAX(r.height(), standardHeight());
	  w = r.width() * h / hh;
	}
	break;
  }
  return w;
}

int InkWpChar::draw(QPainter &p, int x, int baseLine, int h) {
  int w = 0;
  switch(type()) {
  case String: 
	if (text() == " ") {
	  w = h * SPACE_WIDTH;
	} else if (text() == "\n") {
	  w = h * ENTER_WIDTH;
	  p.setPen(QColor(0, 0, 255));
	  p.drawLine(x+w-1, baseLine-h, x+w-1, baseLine-2);
	  p.drawLine(x, baseLine-2, x+w-1, baseLine-2);
	  p.drawLine(x, baseLine-2, x+(w-1)/2, baseLine-(w-1)/2);
	} else if (text() == "") {
	  w = h * END_WIDTH;
	  p.fillRect(x, baseLine - h, w, h, QColor(0, 0, 255));
	}
	break;
  case Handwriting: 
	{
	  QRect r = points().boundingRect();
	  int hh = QMAX(r.height(), standardHeight());
	  w = r.width() * h / hh;

	  int x0 = r.left();
	  int y0 = r.bottom();

	  QPoint p0, p1;
	  p0 = points()[0];
	  for(int i=1; i<points().count(); i++) {
		p1 = points()[i];
		if (connected()[i]) {
		  p.drawLine((p0.x() - x0) * h / hh + x, 
					 (p0.y() - y0) * h / hh + baseLine, 
					 (p1.x() - x0) * h / hh + x, 
					 (p1.y() - y0) * h / hh + baseLine); 
		}
		p0 = p1;
	  }
	}
	break;
  }
  return w;
}

void InkWpChar::addPoint(QPoint point, bool _connected) {
  if (type_ != Handwriting)
	return;
  int s = points_.size();
  points_.resize(s+1);
  connected_.resize(s+1);
  points_[s] = point;
  connected_[s] = _connected;
}

void InkWpChar::setData(const QPointArray &pa, const QBitArray &ba) {
  if (type_ != Handwriting)
	return;
  points_ = pa;
  connected_ = ba;
}
void InkWpChar::setText(const QString &s) {
  if (type_ != String)
	return;
  text_ = s;
}

QDataStream &operator<< (QDataStream &ds, const InkWpChar &c) {
  switch(c.type()) {
  case InkWpChar::Handwriting:
	ds << (Q_UINT8) 'H';

	ds << (Q_UINT8) 'H' << (Q_INT32) c.standardHeight();
	ds << (Q_UINT8) 'P' << c.points();
	ds << (Q_UINT8) 'C' << c.connected();
	ds << (Q_UINT8) 'E';
	break;
  case InkWpChar::String:
	ds << (Q_UINT8) 'S';

	ds << (Q_UINT8) 'T' << c.text();
	ds << (Q_UINT8) 'E';
	break;
  }
  return ds;
}

QDataStream &operator>> (QDataStream &ds, InkWpChar &c) {
  Q_UINT8 i;
  ds >> i;
  switch(i) {
  case 'H':
	{
	  c.setType(InkWpChar::Handwriting);
	  Q_UINT32 h;
	  QPointArray pa;
	  QBitArray ba;
	  while(true) {
		Q_UINT8 i;
		ds >> i;
		if (i == 'H') 
		  ds >> h;
		else if (i == 'P') 
		  ds >> pa;
		else if (i == 'C') 
		  ds >> ba;
		else if (i == 'E')
		  break;
		else
		  break;
	  }
	  c.setStandardHeight(h);
	  c.setData(pa, ba);
	}
	break;
  case 'S':
	{
	  c.setType(InkWpChar::String);
	  QString s;
	  while(true) {
		Q_UINT8 i;
		ds >> i;
		if (i == 'T') 
		  ds >> s;
		else if (i == 'E')
		  break;
		else break;
	  }
	  c.setText(s);
	}
	break;
  default:
    break;
  }
  return ds;
}


InkWpData::InkWpData() {
  clear();
}
bool InkWpData::save(QString fileName) {
  QFile f(fileName);
  if (!f.open(IO_WriteOnly))
	return false;
  QDataStream ds(&f);

  ds << QString(MAGIC) << (Q_UINT8) VER;
  for(QValueListIterator<InkWpChar> it = data_.begin();
	  it != data_.end();
	  ++it) 
	ds << *it;
  f.close();
  return true;
}

bool InkWpData::load(QString fileName) {
  QFile f(fileName);
  if (!f.open(IO_ReadOnly))
	return false;
  QDataStream ds(&f);

  QString magic;
  Q_UINT8 ver;

  ds >> magic >> ver;
  if (magic != MAGIC || ver > VER) {
    clear();
    return false;
  }

  data_.clear();
  while(!f.atEnd()) {
	InkWpChar c;
	ds >> c;
	data_.append(c);
  }
  f.close();
  return true;
}

void InkWpData::clear() {
  data_.clear();
  data_.append(InkWpChar(""));
}

const InkWpChar &InkWpData::at(int i) {
  return *(data_.at(i));
}
int InkWpData::count() const {
  return data_.count();
}
bool InkWpData::remove(int i) {
  if (i < 0 || i >= data_.count()) 
	return false;
  data_.remove(data_.at(i));
  return true;
}
bool InkWpData::insert(int i, const InkWpChar &c) {
  if (i >= data_.count())
	return false;
  if (i < 0) 
	data_.append(c);
  else
	data_.insert(data_.at(i), c);
  return true;
}

#define ROW(y) (((y)+rowHeight)/rowHeight-1)
int InkWpData::pointToIndex(QPoint p, 
							int width, int charHeight, 
							int charMargin, int rowMargin) {
  int x = 0, y = charHeight + rowMargin - 1; 
  int rowHeight = charHeight + rowMargin;

  QValueListIterator<InkWpChar> it = data_.begin();
  int i = 0;
  for(; it != data_.end(); ++it, i++) {
	if (ROW(p.y()) < ROW(y) || 
		(ROW(p.y()) == ROW(y) && p.x() <= x)) 
	  break;
    InkWpChar c = *it;
    int w = c.width(charHeight);
    if (x + w > width) {
      x = 0;
      y += rowHeight;
	  if (ROW(p.y()) < ROW(y) || (ROW(p.y()) == ROW(y) && p.x() <= x))
		break;
	  //	  if (ROW(p.y()) <= ROW(y) && p.x() < 0) 
	  //		break;
    }

    x += w + charMargin;
    if (c.type() == InkWpChar::String && c.text() == "\n") {
	  if (ROW(y) == ROW(p.y()))
		break;
      x = 0;
      y += rowHeight;
	}
  }
  if (i > 0 && i >= data_.count())
	i--;
  return i;
}

QPoint InkWpData::draw(QPainter &p, int currentIndex, 
		     int width, int charHeight, 
		     int charMargin, int rowMargin) {
  int x = 0, y = charHeight + rowMargin - 1;
  QPoint currentPos(x, y);

  QValueListIterator<InkWpChar> it = data_.begin();
  int i = 0;

  p.setPen(QColor(192, 192, 192));
  p.drawLine(0, y, width, y);

  for(; it != data_.end(); ++it, i++) {
    InkWpChar c = *it;
    int w = c.width(charHeight);
    if (x + w > width) {
      x = 0;
      y += charHeight + rowMargin;
      p.setPen(QColor(192, 192, 192));
      p.drawLine(0, y, width, y);
    }
    if (i == currentIndex) {
      p.setPen(QColor(255, 0, 0));
      p.drawLine(x, y-1, x, y - charHeight - rowMargin);
	  currentPos = QPoint(x, y);
    }
    if (i == currentIndex - 1)
      p.setPen(QColor(255, 0, 0));
    else
      p.setPen(QColor(0, 0, 0));
    c.draw(p, x, y, charHeight);

    x += w + charMargin;
    if (c.type() == InkWpChar::String && c.text() == "\n") {
      x = 0;
      y += charHeight + rowMargin;
      p.setPen(QColor(192, 192, 192));
      p.drawLine(0, y, width, y);
	}
  }
  return currentPos;
}
int InkWpData::numberOfLines(int width, int charHeight, 
			     int charMargin, int rowMargin) {
  int x = 0, y = charHeight + rowMargin - 1;
  int count = 1;

  QValueListIterator<InkWpChar> it = data_.begin();

  for(; it != data_.end(); ++it) {
    InkWpChar c = *it;
    int w = c.width(charHeight);
    if (x + w > width) {
      x = 0;
      y += charHeight + rowMargin;
      count++;
    }

    x += w + charMargin;
    if (c.type() == InkWpChar::String && c.text() == "\n") {
      x = 0;
      y += charHeight + rowMargin;
      count++;
	}
  }
  return count;
}
