/*
 * Copyright (C) 2003 ichitokumei, tt. 
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <qwidget.h>
#include <qpainter.h>
#include <qtimer.h>
#include <qapplication.h>

#include "inkwpwidget.h"
#include "inkwpdata.h"

// for input
#define WAIT 500

InkWpWidget::InkWpWidget(QWidget* parent,  const char* name, 
			 InkWpData *data, bool readOnly)
  : QWidget(parent, name), data_(data), readOnly_(readOnly)
{
  if (readOnly_)
    setFocusPolicy(NoFocus);
  else
    setFocusPolicy(StrongFocus);
  setBackgroundColor(QColor(255, 255, 255));

  currentIndex_ = 0;
  waitForPoint_ = false;
  twoLineMode_ = false;
  charHeight_ = 12;

  timer = new QTimer();
  connect(timer, SIGNAL(timeout()), this, SLOT(slotTimeOut()));
}

void InkWpWidget::mousePressEvent(QMouseEvent* e) 
{
  if (readOnly_)
	return;
  if (waitForPoint_)
    return;

  if (twoLineMode_) {
    if ((width() < height() && 
	 timer->isActive() &&
	 (e->pos().y() - (height()/2)) * (lastY - (height()/2)) <= 0) ||
	(width() >= height() && 
	 timer->isActive() &&
	 (e->pos().x() - (width()/2)) * (lastX - (width()/2)) <= 0)) {
      timer->stop();
      slotTimeOut();
    }
  }

  lastX = e->pos().x();
  lastY = e->pos().y();
  timer->stop();
  p0 = e->pos();
  p1 = p0;

  int s = points.size();
  points.resize(s + 1);
  connected.resize(s + 1);
  points[s] = p0;
  connected[s] = false;
}

void InkWpWidget::jumpTo(QPoint q) {
  currentIndex_ = data_->pointToIndex(q, width(), charHeight_, 
				      charMargin_, rowMargin_);
  repaint();
}

void InkWpWidget::mouseReleaseEvent(QMouseEvent* e){
  if (readOnly_)
	return;
  if (waitForPoint_) {
	waitForPoint_ = false;
    jumpTo(e->pos());
	emit pointSelected();
    return;
  }

  int s = points.size();
  points.resize(s + 1);
  connected.resize(s + 1);
  points[s] = e->pos();
  connected[s] = true;

  timer->start(WAIT, true);
}

void InkWpWidget::mouseMoveEvent(QMouseEvent* e){
  if (readOnly_)
	return;
  if (waitForPoint_)
    return;

  QPoint p = e->pos();

  int s = points.size();
  QWidget *w = QApplication::desktop();
  // FIXME: should make configurable?
  int threshold = QMAX(w->width(), w->height()) / 16; 
  if ((p.x() - p0.x() < 0 && p0.x() - p1.x() >= 0) || 
      (p.x() - p0.x() > 0 && p0.x() - p1.x() <= 0) || 
      (p.y() - p0.y() < 0 && p0.y() - p1.y() >= 0) || 
      (p.y() - p0.y() > 0 && p0.y() - p1.y() <= 0) || // curved in some way
      (p - points[s-1]).manhattanLength() >= threshold) { 
	points.resize(s + 1);
	connected.resize(s + 1);
	points[s] = p;
	connected[s] = true;
  }

  QPainter painter(this);
  painter.setPen(red);
  painter.drawLine(p0, p);
  painter.end();

  p1 = p0;
  p0 = p;
}

void InkWpWidget::keyPressEvent(QKeyEvent* e){
  switch(e->key()) {
  case Qt::Key_Left:
    if (currentIndex_ > 0)
      currentIndex_--;
    repaint();
    break;
  case Qt::Key_Right:
    if (currentIndex_ < data_->count() - 1)
      currentIndex_++;
    repaint();
    break;
  case Qt::Key_Space:
    slotSpace();
    break;
  case Qt::Key_Backspace:
    slotBS();
    break;
  case Qt::Key_Return:
    slotEnter();
    break;
  case Qt::Key_Up:
	{
	  int savedX = currentPos_.x();
	  jumpTo(currentPos_ + QPoint(-1, -(charHeight_ + rowMargin_)));
	  if (currentIndex_ != 0)
		currentPos_.setX(savedX);
	}
    break;
  case Qt::Key_Down:
	{
	  int savedX = currentPos_.x();
	  jumpTo(currentPos_ + QPoint(-1, charHeight_ + rowMargin_));
	  if (currentIndex_ != 0)
		currentPos_.setX(savedX);
	}
    break;
  }
}

void InkWpWidget::slotTimeOut() {
  InkWpChar c(height()/3); // FIXME: should make configurable ?
  c.setData(points, connected);
  dirty_ = true;
  data_->insert(currentIndex_, c);
  currentIndex_++;
  //  if (countLines() > ROWS) {
  //    data.remove();
  //  } else {
  //  dirty_ = true;
  //  currentIndex_++;
  //  }

  repaint();

  points.resize(0);
  connected.resize(0);
}

void InkWpWidget::paintEvent(QPaintEvent*) {
  QPainter painter(this);
  currentPos_ = data_->draw(painter, currentIndex_, 
			    width(), charHeight_, charMargin_, rowMargin_);
  if (!readOnly_) {
    if (twoLineMode_) {
      painter.setPen(QColor(192, 192, 255));
      if (width() < height()) 
	painter.drawLine(0, height()/2, width(), height()/2);
      else
	painter.drawLine(width()/2, 0, width()/2, height());
    }
  }
}

void InkWpWidget::slotSpace() {
  if (!isVisible())
    return;
  if (timer->isActive()) {
    timer->stop();
    slotTimeOut();
  }

  InkWpChar c(" ");
  dirty_ = true;
  data_->insert(currentIndex_, c);
  currentIndex_++;
  /*
  if (countLines() > ROWS) {
    data.remove();
  } else {
    dirty_ = true;
    data.next();
  }
  */
  repaint();
}
void InkWpWidget::slotBS() {
  if (!isVisible())
    return;
  if (timer->isActive()) {
    timer->stop();
    slotTimeOut();
  }

  if (currentIndex_ > 0) {
    currentIndex_--;
    data_->remove(currentIndex_);
  }
  repaint();
}
void InkWpWidget::slotDeleteLine() {
  if (!isVisible())
    return;
  if (timer->isActive()) {
    timer->stop();
    slotTimeOut();
  }

  int p = currentIndex_;
  for (;currentIndex_ > 0;) {
    const InkWpChar& c = data_->at(currentIndex_-1);
    if (c.type() == InkWpChar::String && c.text() == "\n") {
      if (currentIndex_ == p) {
        currentIndex_--;
        data_->remove(currentIndex_);
      }
      break;
    } else {
      currentIndex_--;
      data_->remove(currentIndex_);
    }
  }
  repaint();
}
void InkWpWidget::slotEnter() {
  if (!isVisible())
    return;
  if (timer->isActive()) {
    timer->stop();
    slotTimeOut();
  }

  InkWpChar c("\n");
  dirty_ = true;
  data_->insert(currentIndex_, c);
  currentIndex_++;
  /*
  if (countLines() > ROWS) {
    data.remove();
  } else {
    dirty = true;
    data.next();
  }
  */
  repaint();
}

void InkWpWidget::setTwoLineMode(bool b) {
  twoLineMode_ = b;
  repaint();
}

void InkWpWidget::setDimensions(int charHeight, 
				int charMargin, int rowMargin) {
  charHeight_ = charHeight;
  charMargin_ = charMargin;
  rowMargin_ = rowMargin;
  repaint();
}
